use std::{error::Error, path::PathBuf, sync::Arc, time::Duration};

use base64::{engine::general_purpose, Engine};
use colored::Colorize;
use futures::{lock::Mutex, stream, StreamExt};
use md5::{Digest, Md5};
use reqwest::{header::USER_AGENT, Client};
use tokio::{fs::File, io::AsyncWriteExt};

use crate::UA;

const KEEP_ALIVE_TIMEOUT: Duration = Duration::from_secs(5);

pub async fn concurrent_dl(
    images: Vec<(String, PathBuf, String)>,
) -> Result<(usize, usize), Box<dyn Error>> {
    let dl_count = Arc::new(Mutex::new(0));
    let sk_count = Arc::new(Mutex::new(0));

    let client = Client::builder()
        .pool_idle_timeout(KEEP_ALIVE_TIMEOUT)
        .build()?;

    let futures = stream::iter(images.iter().map(|data| async {
        let dl_count = Arc::clone(&dl_count);
        let _sk_count = Arc::clone(&sk_count);
        let client = client.clone();

        let (url, path, _expct_md5) = data;
        let send_fut = client.get(url).header(USER_AGENT, UA).send();

        match send_fut.await {
            Ok(res) => match res.bytes().await {
                Ok(bytes) => {
                    let byte_count = bytes.len();

                    let mut hasher = Md5::new();
                    hasher.update(&bytes);
                    let result = hasher.finalize();
                    let b64_md5 = general_purpose::STANDARD.encode(result);

                    // 4chan file attachment hash is always 24 character packed base64 encoded MD5. Truly a fucking state of art CRC.

                    // TODO: Figure out how the MD5 should be converted before uncommenting the following filtering condition

                    // if b64_md5 != *expct_md5 {
                    //     eprintln!("{}", format!("File skipped due to mismatched MD5 (expected {expct_md5}, got {b64_md5})").red().bold());
                    //     let mut sk_count = sk_count.lock().await;
                    //     *sk_count += 1;

                    //     return;
                    // }

                    let mut file = File::create(path).await.unwrap();
                    file.write_all(&bytes).await.unwrap();

                    let mut dl_count = dl_count.lock().await;
                    *dl_count += 1;

                    println!(
                        "{}",
                        format!("{b64_md5}: {byte_count} bytes").truecolor(0, 209, 27)
                    );
                }
                Err(_) => eprintln!(
                    "{}",
                    format!("Failed to convert request from {} to bytes", url)
                        .red()
                        .bold()
                ),
            },
            Err(_) => eprintln!("{}", format!("Failed to request {}", url).red().bold()),
        }
    }))
    .buffer_unordered(100)
    .collect::<Vec<()>>();

    futures.await;

    let dl = *dl_count.lock().await;
    let sk = *sk_count.lock().await;

    Ok((dl, sk))
}
